# -*- coding: utf-8 -*-

from plone.dexterity.browser.edit import DefaultEditForm
from z3c.form.interfaces import DISPLAY_MODE
from plone.z3cform import layout


class CustomViewForm(DefaultEditForm):
    mode = DISPLAY_MODE

    _actions_keys = ['save', 'cancel']

    def update(self):
        super(CustomViewForm, self).update()
        for action in self._actions_keys:
            del self.actions[action]


CustomView = layout.wrap_form(CustomViewForm)



# -*- coding: utf-8 -*-

from collective.task import _ as CTMF
from collective.task.behaviors import ITask
from collective.task.behaviors import get_parent_assigned_group
from collective.task.behaviors import get_users_vocabulary
from collective.task.field import LocalRoleMasterSelectField

from plone import api
from plone.autoform import directives
from plone.dexterity.content import Container
from plone.dexterity.content import Item

from imio.schedule.config import CREATION
from imio.schedule.config import DONE
from imio.schedule.config import STARTED
from imio.schedule.config import status_by_state
from imio.schedule.interfaces import ScheduleConfigNotFound
from imio.schedule.interfaces import TaskConfigNotFound

from plone.memoize.request import cache

from zope.interface import implements


class IAutomatedTask(ITask):
    """
    AutomatedTask dexterity schema.
    """

    directives.order_before(assigned_group='assigned_user')
    directives.order_before(assigned_group='ITask.assigned_user')
    assigned_group = LocalRoleMasterSelectField(
        title=CTMF(u"Assigned group"),
        required=True,
        vocabulary="collective.task.AssignedGroups",
        slave_fields=(
            {'name': 'ITask.assigned_user',
             'slaveID': '#form-widgets-ITask-assigned_user',
             'action': 'vocabulary',
             'vocab_method': get_users_vocabulary,
             'control_param': 'group',
             },
        ),
        defaultFactory=get_parent_assigned_group
    )


class BaseAutomatedTask(object):
    """
    Base class for AutomatedTask content types.
    """

    task_config_UID = ''
    schedule_config_UID = ''

    def get_container(self):
        """
        Return the task container.
        """
        container = self
        while IAutomatedTask.providedBy(container):
            container = container.getParentNode()

        return container

    def level(self):
        """
        Return the task depth contenance level.
        """
        container = self
        level = -1
        while IAutomatedTask.providedBy(container):
            container = container.getParentNode()
            level = level + 1

        return level

    def get_schedule_config(self):
        """
        Return associated schedule config.
        """
        catalog = api.portal.get_tool('portal_catalog')
        brains = catalog(UID=self.schedule_config_UID)
        if brains:
            return brains[0].getObject()
        else:
            raise ScheduleConfigNotFound(
                'UID {}'.format(self.schedule_config_UID)
            )

    @cache(get_key=lambda m, task: task.id, get_request='self.REQUEST')
    def get_task_config(self):
        """
        Return associated task config.
        """
        catalog = api.portal.get_tool('portal_catalog')
        with api.env.adopt_roles(['Manager']):
            brains = catalog.unrestrictedSearchResults(UID=self.task_config_UID)
            if brains:
                return brains[0].getObject()
            else:
                raise TaskConfigNotFound(
                    'UID {}'.format(self.task_config_UID)
                )

    def get_status(self):
        """
        Return the status of the task
        """
        return status_by_state[self.get_state()]

    @cache(get_key=lambda m, task: task.id, get_request='self.REQUEST')
    def start_conditions_status(self):
        """
        See start_conditions_status of TaskConfig.
        """
        task_config = self.get_task_config()
        container = self.get_container()
        status = task_config.start_conditions_status(container, self)
        return status

    @cache(get_key=lambda m, task: task.id, get_request='self.REQUEST')
    def starting_states_status(self):
        """
        """
        config = self.get_task_config()
        starting_states = config.starting_states
        if not starting_states:
            return

        container = self.get_container()
        container_state = api.content.get_state(container)
        return (container_state, starting_states)

    @cache(get_key=lambda m, task: task.id, get_request='self.REQUEST')
    def end_conditions_status(self):
        """
        See end_conditions_status of TaskConfig.
        """
        task_config = self.get_task_config()
        container = self.get_container()
        status = task_config.end_conditions_status(container, self)
        return status

    def ending_states_status(self):
        """
        """
        config = self.get_task_config()
        ending_states = config.ending_states
        if not ending_states:
            return

        container = self.get_container()
        container_state = api.content.get_state(container)
        return (container_state, ending_states)

    def get_state(self):
        return api.content.get_state(self)

    def get_subtasks(self):
        """
        A normal task has no sub tasks.
        """
        return []

    @property
    def end_date(self):
        """
        """
        if self.get_status() == DONE:
            wf_history = self.workflow_history['task_workflow'][::-1]
            for action in wf_history:
                if status_by_state[action['review_state']] is DONE:
                    end_date = action['time']
                    return end_date.asdatetime()
        return None

    def reindex_parent_tasks(self, idxs=[]):
        """
        Reindex 'idxs' indexes of all the parent tasks of this task.
        """
        # reindex parent tasks
        parent_task = self.getParentNode()
        while IAutomatedTask.providedBy(parent_task):
            parent_task.reindexObject(idxs)
            parent_task = parent_task.getParentNode()

    @property
    def is_solvable(self):
        """
        Return True if this task and its OPEN (sub-)subtasks have the same assigned_user.
        """
        subtasks = self.get_subtasks()
        while subtasks:
            subtask = subtasks.pop()
            if subtask.get_status() in [CREATION, STARTED]:
                if subtask.assigned_user != self.assigned_user:
                    return False
                subtasks.extend(subtask.get_subtasks())

        return True

    def _start(self):
        """
        Delegate start operation to the task_config
        """
        task_config = self.get_task_config()
        task_config.start_task(self)

    def _end(self):
        """
        Delegate end operation to the task_config
        """
        task_config = self.get_task_config()
        with api.env.adopt_roles(['Manager']):
            task_config.end_task(self)

    def _freeze(self):
        """
        Delegate freeze operation to the task_config
        """
        task_config = self.get_task_config()
        task_config.freeze_task(self)

    def _thaw(self):
        """
        Delegate thaw operation to the task_config
        """
        task_config = self.get_task_config()
        task_config.thaw_task(self)


class AutomatedTask(Item, BaseAutomatedTask):
    """
    """

    implements(IAutomatedTask)


class IAutomatedMacroTask(IAutomatedTask):
    """
    AutomatedTask dexterity schema.
    """


class AutomatedMacroTask(Container, BaseAutomatedTask):
    """
    """

    implements(IAutomatedMacroTask)

    def get_subtasks(self):
        """
        Return all sub tasks of this macro task.
        """
        sub_tasks = [obj for obj in self.objectValues() if ITask.providedBy(obj)]
        return sub_tasks

    def get_last_subtasks(self):
        """
        Return each last unique sub task of this macro task.
        """
        subtask_type = set()
        sub_tasks = []

        for obj in reversed(self.objectValues()):
            if ITask.providedBy(obj):
                subtask = obj
                if subtask.task_config_UID not in subtask_type:
                    subtask_type.add(subtask.task_config_UID)
                    sub_tasks.append(subtask)

        return reversed(sub_tasks)



# -*- coding: utf-8 -*-

from plone import api
from zope.interface import alsoProvides


def upgrade_2_fix_schedule_collection(context):
    from collective.eeafaceted.collectionwidget.interfaces import IDashboardCollection
    from imio.schedule.interfaces import IScheduleCollection

    brains = api.content.find(
        object_provides=IDashboardCollection.__identifier__,
        id="dashboard_collection",
    )
    for brain in brains:
        if "schedule" not in brain.getPath():
            continue
        collection = brain.getObject()
        alsoProvides(collection, IScheduleCollection)


def upgrade_3_set_showNumberOfItems(context):
    from collective.eeafaceted.collectionwidget.interfaces import IDashboardCollection
    brains = api.content.find(
        object_provides=IDashboardCollection.__identifier__,
        id="dashboard_collection",
    )
    for brain in brains:
        if "schedule" not in brain.getPath():
            continue
        collection = brain.getObject()
        collection.showNumberOfItems = True



# -*- coding: utf-8 -*-
"""Module where all interfaces, events and exceptions live."""

from z3c.table.interfaces import IColumn

from zope.interface import Interface


class IDisplayTaskStatus(Interface):
    """
    Adapts a task instance into a z3c table cell
    displaying the task status.
    """

    def render(self):
        """
        """


class ISimpleDisplayTaskStatus(Interface):
    """
    Adapts a task instance into a z3c table cell
    displaying the task status.
    """

    def render(self):
        """
        """


class IStatusColumn(IColumn):
    """ Marker interface for Task Status columns."""



# -*- coding: utf-8 -*-

from plone import api

from Products.ATContentTypes.interfaces import IATFolder
from Products.ATContentTypes.interfaces import IATFile

from imio.schedule.utils import interface_to_tuple
from imio.schedule.content.object_factories import CreationConditionObject
from imio.schedule.content.object_factories import StartConditionObject
from imio.schedule.content.object_factories import EndConditionObject
from imio.schedule.content.object_factories import FreezeConditionObject
from imio.schedule.content.object_factories import MacroCreationConditionObject
from imio.schedule.content.object_factories import MacroStartConditionObject
from imio.schedule.content.object_factories import MacroEndConditionObject
from imio.schedule.content.object_factories import ThawConditionObject


def schedule_example_install(context):
    """
    Example schedule install script.
    """
    if context.readDataFile('imioschedule_testing.txt') is None:
        return

    monkey_patch_scheduled_conttentype_vocabulary(context)
    add_empty_task_container(context)
    add_schedule_config(context)
    add_task(context)


def monkey_patch_scheduled_conttentype_vocabulary(context):
    """
    Monkey patch the default vocabulary for the field 'scheduled_contenttype'
    to be able to select 'Folder' content type so we can test methods of this
    field.
    """
    from imio.schedule.content.vocabulary import ScheduledContentTypeVocabulary

    def monkey_allowed_types(self):
        return{'Folder': IATFolder}
    ScheduledContentTypeVocabulary.content_types = monkey_allowed_types


def add_empty_task_container(context):
    """
    Add dummy empty task container (ATFolder)
    """
    site = api.portal.get()

    folder_id = 'test_empty_taskcontainer'
    if folder_id not in site.objectIds():
        api.content.create(
            container=site,
            type='Folder',
            id=folder_id,
            title='Empty Task container'
        )


def add_schedule_config(context):
    """
    Add dummy ScheduleConfig, TaskConfig.
    """
    site = api.portal.get()

    # create config folder for schedule config
    folder_id = 'config'
    if folder_id not in site.objectIds():
        api.content.create(
            container=site,
            type='Folder',
            id=folder_id,
            title='Task configs'
        )
    cfg_folder = getattr(site, folder_id)

    # create empty schedule config
    schedule_cfg_id = 'empty_scheduleconfig'
    if schedule_cfg_id not in cfg_folder.objectIds():
        api.content.create(
            container=cfg_folder,
            type='ScheduleConfig',
            id=schedule_cfg_id,
            title='Empty ScheduleConfig',
            scheduled_contenttype=('Folder', (interface_to_tuple(IATFile),)),
        )

    # create schedule config
    schedule_cfg_id = 'test_scheduleconfig'
    if schedule_cfg_id not in cfg_folder.objectIds():
        api.content.create(
            container=cfg_folder,
            type='ScheduleConfig',
            id=schedule_cfg_id,
            title='Test ScheduleConfig',
            scheduled_contenttype=('Folder', (interface_to_tuple(IATFolder),)),
        )
    schedule_config = getattr(cfg_folder, schedule_cfg_id)

    # create task config
    task_cfg_id = 'test_taskconfig'
    creation_conditions = CreationConditionObject()
    creation_conditions.__dict__ = {
        'condition': u'schedule.test_creation_condition',
        'operator': 'AND',
    }
    start_conditions = StartConditionObject()
    start_conditions.__dict__ = {
        'condition': u'schedule.test_start_condition',
        'operator': 'AND',
    }
    end_conditions = EndConditionObject()
    end_conditions.__dict__ = {
        'condition': u'schedule.test_end_condition',
        'operator': 'AND',
    }
    freeze_conditions = FreezeConditionObject()
    freeze_conditions.__dict__ = {
        'condition': u'schedule.test_freeze_condition',
        'operator': 'AND',
    }
    thaw_conditions = ThawConditionObject()
    thaw_conditions.__dict__ = {
        'condition': u'schedule.test_thaw_condition',
        'operator': 'AND',
    }
    macro_creation_conditions = MacroCreationConditionObject()
    macro_creation_conditions.__dict__ = creation_conditions.__dict__
    macro_start_conditions = MacroStartConditionObject()
    macro_start_conditions.__dict__ = start_conditions.__dict__
    macro_end_conditions = MacroEndConditionObject()
    macro_end_conditions.__dict__ = end_conditions.__dict__

    if task_cfg_id not in schedule_config.objectIds():
        api.content.create(
            container=schedule_config,
            type='TaskConfig',
            id=task_cfg_id,
            title='Test TaskConfig',
            default_assigned_user='schedule.assign_current_user',
            default_assigned_group='schedule.assign_authenticatedusers_group',
            creation_conditions=[creation_conditions],
            start_conditions=[start_conditions],
            end_conditions=[end_conditions],
            freeze_conditions=[freeze_conditions],
            thaw_conditions=[thaw_conditions],
            creation_state='private',
            starting_states=('pending',),
            ending_states=('published',),
            start_date='schedule.start_date.creation_date',
            additional_delay=10,
        )

    # create macro task config
    macrotask_cfg_id = 'test_macrotaskconfig'
    if macrotask_cfg_id not in schedule_config.objectIds():
        api.content.create(
            container=schedule_config,
            type='MacroTaskConfig',
            id=macrotask_cfg_id,
            title='Test MacroTaskConfig',
            default_assigned_user='schedule.assign_current_user',
            default_assigned_group='schedule.assign_authenticatedusers_group',
            creation_conditions=[macro_creation_conditions],
            start_conditions=[macro_start_conditions],
            end_conditions=[macro_end_conditions],
            creation_state='private',
            starting_states=('pending',),
            ending_states=('published',),
            start_date='schedule.start_date.subtask_highest_due_date',
            additional_delay=17,
        )
    macrotask_config = getattr(schedule_config, macrotask_cfg_id)

    # create sub task config
    subtask_cfg_id = 'test_subtaskconfig'
    if subtask_cfg_id not in macrotask_config.objectIds():
        api.content.create(
            container=macrotask_config,
            type='TaskConfig',
            id=subtask_cfg_id,
            title='Test SubTaskConfig',
            default_assigned_user='schedule.assign_current_user',
            default_assigned_group='schedule.assign_authenticatedusers_group',
            creation_conditions=[creation_conditions],
            start_conditions=[start_conditions],
            end_conditions=[end_conditions],
            creation_state='private',
            starting_states=('pending',),
            ending_states=('published',),
            start_date='schedule.start_date.creation_date',
            additional_delay=13,
        )


def add_task(context):
    """
    Add dummy task container (ATFolder) and create
    a AutomatedTask in it
    """
    site = api.portal.get()

    task_container_id = 'test_taskcontainer'
    if task_container_id not in site.objectIds():
        api.content.create(
            container=site,
            type='Folder',
            id=task_container_id,
            title='Task container'
        )
    task_container = getattr(site, task_container_id)

    # If no task was created automatically, create the task manually
    # to keep AutomatedTask tests alive
    task_id = 'TASK_test_taskconfig'
    if task_id not in task_container.objectIds():
        portal_types = api.portal.get_tool('portal_types')
        type_info = portal_types.getTypeInfo('AutomatedTask')
        schedule_config = site.config.test_scheduleconfig
        task_config = schedule_config.test_taskconfig

        type_info._constructInstance(
            container=task_container,
            id=task_id,
            title=task_config.Title(),
            schedule_config_UID=schedule_config.UID(),
            task_config_UID=task_config.UID()
        )

    # If no task was created automatically, create the macrotask manually
    # to keep AutomatedMacroTask tests alive
    macrotask_id = 'TASK_test_macrotaskconfig'
    if macrotask_id not in task_container.objectIds():
        portal_types = api.portal.get_tool('portal_types')
        type_info = portal_types.getTypeInfo('AutomatedMacroTask')
        schedule_config = site.config.test_scheduleconfig
        task_config = schedule_config.test_macrotaskconfig

        type_info._constructInstance(
            container=task_container,
            id=macrotask_id,
            title=task_config.Title(),
            schedule_config_UID=schedule_config.UID(),
            task_config_UID=task_config.UID()
        )
    macro_task = getattr(task_container, macrotask_id)

    # If no task was created automatically, create the subtask manually
    # to keep AutomatedMacroTask tests alive
    subtask_id = 'TASK_test_subtaskconfig'
    if subtask_id not in macro_task.objectIds():
        portal_types = api.portal.get_tool('portal_types')
        type_info = portal_types.getTypeInfo('AutomatedTask')
        schedule_config = site.config.test_scheduleconfig
        task_config = schedule_config.test_macrotaskconfig.test_subtaskconfig

        type_info._constructInstance(
            container=macro_task,
            id=subtask_id,
            title=task_config.Title(),
            schedule_config_UID=schedule_config.UID(),
            task_config_UID=task_config.UID()
        )



# -*- coding: utf-8 -*-

from imio.schedule import _
from imio.schedule.content.schedule_config import IScheduleConfig
from imio.schedule.interfaces import IScheduleCollection

from plone import api

from zope.interface import alsoProvides


def create(schedule_container, event):
    """
    """
    #do not automatically re-create the collection during upgrade steps
    if 'portal_setup/manage_importSteps' in schedule_container.REQUEST.URL:
        return

    collection_id = 'dashboard_collection'
    title = IScheduleConfig.providedBy(schedule_container) and _('All') or schedule_container.Title()

    if collection_id not in schedule_container.objectIds():
        factory_args = {
            'id': 'dashboard_collection',
            'title': title,
            'query': [
                {
                    'i': 'CompoundCriterion',
                    'o': 'plone.app.querystring.operation.compound.is',
                    'v': schedule_container.UID()
                },
                {
                    'i': 'review_state',
                    'o': 'plone.app.querystring.operation.selection.is',
                    'v': ['to_do']
                }
            ],
            'customViewFields': (
                u'assigned_user',
                u'status',
                u'due_date'
            ),
            'sort_on': u'due_date',
            'sort_reversed': True,
            'b_size': 30
        }

        if IScheduleConfig.providedBy(schedule_container):
            factory_args['customViewFields'] = (
                u'pretty_link',
                u'sortable_title',
                u'assigned_user',
                u'status',
                u'due_date'
            )

        kwargs = {}
        additional_queries = kwargs.pop('query', [])
        for query in additional_queries:
            factory_args['query'].append(query)
        factory_args.update(kwargs)

        portal_types = api.portal.get_tool('portal_types')
        type_info = portal_types.getTypeInfo('DashboardCollection')
        collection = type_info._constructInstance(schedule_container, **factory_args)
        # mark the collection with an interface to to customize the render
        # term view of collection widget
        alsoProvides(collection, IScheduleCollection)


def update_title(schedule_container, event):
    """
    Dashboard Collection title should always be the title of the parent task.
    """
    # do not automatically re-create the collection during upgrade steps
    if 'portal_setup' in schedule_container.REQUEST.URL:
        return

    collection = getattr(schedule_container, 'dashboard_collection', None)
    if collection:
        title = IScheduleConfig.providedBy(schedule_container) and _('All') or schedule_container.Title()
        collection.title = title
        collection.reindexObject(idxs=('Title', 'sortable_title'))



# -*- coding: utf-8 -*-

from z3c.form import term
from z3c.form.browser.select import SelectWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import IFormLayer
from z3c.form.widget import FieldWidget
from zope import schema
from zope.component import adapter
from zope.component import getUtility
from zope.interface import Interface
from zope.interface import implementer
from zope.interface import implements
from zope.schema.interfaces import IVocabularyFactory


class ISubFormContextChoice(Interface):
    pass


class SubFormContextChoice(schema.Choice):
    implements(ISubFormContextChoice)

    def __init__(self, *args, **kwargs):
        kwargs['values'] = []
        vocabulary_name = kwargs.pop('vocabulary')
        super(SubFormContextChoice, self).__init__(*args, **kwargs)
        self.vocabulary_name = vocabulary_name


class SubFormContextSelectWidget(SelectWidget):

    def update(self):
        super(SubFormContextSelectWidget, self).update()
        self.field.vocabulary = self.terms.terms


@adapter(ISubFormContextChoice, IFormLayer)
@implementer(IFieldWidget)
def subform_context_select_field_widget(field, request):
    return FieldWidget(field, SubFormContextSelectWidget(request))


def subform_context_choice_terms(context, request, form, field, widget):
    if hasattr(form, 'parentForm'):
        context = form.parentForm.context
    field = field.bind(context)
    voc = getUtility(IVocabularyFactory, name=field.vocabulary_name)
    return term.ChoiceTermsVocabulary(
        context,
        request,
        form,
        field,
        voc(context),
        widget,
    )



# -*- coding: utf-8 -*-
"""Module where all interfaces, events and exceptions live."""

from plone.supermodel import model
from zope import schema
from zope.interface import Interface
from zope.interface.interfaces import IInterface
from zope.publisher.interfaces.browser import IDefaultBrowserLayer

from imio.schedule import _


class IImioScheduleLayer(IDefaultBrowserLayer):
    """Marker interface that defines a browser layer."""


class IScheduledContentTypeVocabulary(Interface):
    """
    Adapts a ScheduleConfig instance into a vocabulary.
    """


class ITaskLogic(Interface):
    """
    Base interface for the following TaskConfig logic items:
        - conditions,
        - date computation
        - user assignment
    """


class IDefaultEndingStates(Interface):
    """
    Adapts a TaskContainer into a list of default ending states.
    """


class IDefaultFreezeStates(Interface):
    """
    Adapts a TaskContainer into a list of default freeze states.
    """


class IDefaultThawStates(Interface):
    """
    Adapts a TaskContainer into a list of default freeze states.
    """


class IDefaultTaskGroup(ITaskLogic):
    """
    Adapts a TaskContainer into a plone group to assign to a task.
    """


class IDefaultTaskUser(ITaskLogic):
    """
    Adapts a TaskContainer into a plone user to assign to a task.
    """


class ICondition(ITaskLogic):
    """
    Condition object adapting a TaskContainer and task.
    """

    def evaluate(self):
        """
        evaluate if the condition is True or False
        """


class ICreationCondition(ICondition):
    """
    Creation condition of task.
    """


class IStartCondition(ICondition):
    """
    Start condition of task.
    """


class IEndCondition(ICondition):
    """
    End condition of task.
    """


class IFreezeCondition(ICondition):
    """
    Freeze condition of task.
    """


class IThawCondition(ICondition):
    """
    Thaw condition of task.
    """


class IStartDate(ITaskLogic):
    """
    Adapts a TaskContainer into the start date used to compute
    the task due date.
    """


class IMacroTaskCreationCondition(ICreationCondition):
    """
    Creation condition of macro task.
    """


class IMacroTaskStartCondition(IStartCondition):
    """
    Start condition of macro task.
    """


class IMacroTaskEndCondition(IEndCondition):
    """
    End condition of macro task.
    """


class IMacroTaskFreezeCondition(IFreezeCondition):
    """
    Freeze condition of macro task.
    """


class IMacroTaskThawCondition(IThawCondition):
    """
    Thaw condition of macro task.
    """


class IMacroTaskStartDate(IStartDate):
    """
    Adapts a TaskContainer into the start date used to compute
    the macro task due date.
    """


class IToTaskConfig(Interface):
    """
    Interface for adapters returning the task config of
    a context providing ITaskContainer.
    """


class IScheduleView(Interface):
    """
    Marker interface to provides on folderish with the schedule faceted view.
    """


class IScheduleCollection(Interface):
    """
    Marker interface for collections associated to schedule/task config.
    """


class ITaskMarkerInterface(IInterface):
    """
    Marker interface for for AutomatedTask custom marker interfaces vocabulary.
    """


class ScheduleConfigNotFound(Exception):
    """
    Raised when a ScheduleConfig is not found.
    """


class TaskConfigNotFound(Exception):
    """
    Raised when a TaskConfig is not found.
    """


class TaskAlreadyExists(Exception):
    """
    Raised when a Task already exists.
    """


class IRecurrenceCondition(ITaskLogic):
    """
    Base interface for the TaskConfig recurrence conditions
    """


class ICalculationDelay(ITaskLogic):
    """
    Base interface for the TaskConfig calculation delay methods
    """


class IFreezeDuration(Interface):
    """
    Base interface for freeze duration adapters.
    """


class IToIcon(Interface):
    """
    Adapts an object into an icon.
    """

    def get_icon_url(self):
        """
        Return the icon url.
        """


class ISettings(model.Schema):

    working_days = schema.List(
        title=_(u'List of working days'),
        value_type=schema.Choice(
            title=_(u'Day'),
            vocabulary='imio.schedule.working_days',
        ),
        required=True,
    )


class ICalendarExtraHolidays(Interface):
    """Interface for extra holidays utility"""

    def get_holidays(self, year):
        """
        Return a tuple with the holidays for the given year
        format: ((<date object>, 'holiday_name'), ...)
        """



# -*- coding: utf-8 -*-

from plone import api
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import login
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.testing import z2

from imio.schedule.events import zope_registration
from imio.schedule.events.zope_registration import unsubscribe_task_configs_for_content_type

import transaction

import unittest

import imio.schedule


class NakedPloneLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        """Set up Zope."""
        # Load ZCML
        self.loadZCML(package=imio.schedule,
                      name='testing.zcml')
        # need to do this for archetypes products having 'initialize'
        # method for their content types in their __init__.py
        z2.installProduct(app, 'imio.dashboard')

    def tearDownZope(self, app):
        """Tear down Zope."""
        z2.uninstallProduct(app, 'imio.dashboard')

NAKED_PLONE_FIXTURE = NakedPloneLayer(
    name='NAKED_PLONE_FIXTURE'
)

NAKED_PLONE_INTEGRATION = IntegrationTesting(
    bases=(NAKED_PLONE_FIXTURE,),
    name='NAKED_PLONE_INTEGRATION'
)


class ScheduleLayer(NakedPloneLayer):

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'Products.CMFPlone:plone')
        applyProfile(portal, 'imio.schedule:default')

        # Login and create some test content
        setRoles(portal, TEST_USER_ID, ['Manager'])
        login(portal, TEST_USER_NAME)

        # Commit so that the test browser sees these objects
        transaction.commit()


TEST_INSTALL_FIXTURE = ScheduleLayer(
    name='TEST_INSTALL_FIXTURE'
)

TEST_INSTALL_INTEGRATION = IntegrationTesting(
    bases=(TEST_INSTALL_FIXTURE,),
    name='TEST_INSTALL_INTEGRATION'
)


TEST_INSTALL_FUNCTIONAL = FunctionalTesting(
    bases=(TEST_INSTALL_FIXTURE,),
    name='TEST_INSTALL_FUNCTIONAL'
)


class ExampleScheduleLayer(ScheduleLayer):

    def setUpPloneSite(self, portal):
        super(ExampleScheduleLayer, self).setUpPloneSite(portal)

        applyProfile(portal, 'imio.schedule:testing')

        # delete macro tasks
        api.content.delete(
            portal.config.test_scheduleconfig.test_macrotaskconfig
        )
        api.content.delete(
            portal.test_taskcontainer.TASK_test_macrotaskconfig
        )


EXAMPLE_SCHEDULE_FIXTURE = ExampleScheduleLayer(
    name='EXAMPLE_SCHEDULE_FIXTURE'
)

EXAMPLE_SCHEDULE_INTEGRATION = IntegrationTesting(
    bases=(EXAMPLE_SCHEDULE_FIXTURE,),
    name='EXAMPLE_SCHEDULE_INTEGRATION'
)


EXAMPLE_SCHEDULE_FUNCTIONAL = FunctionalTesting(
    bases=(EXAMPLE_SCHEDULE_FIXTURE,),
    name='EXAMPLE_SCHEDULE_FUNCTIONAL'
)


class BaseTest(unittest.TestCase):
    """
    Helper class for tests.
    """

    def setUp(self):
        self.portal = self.layer['portal']


class BrowserTest(BaseTest):
    """
    Helper class for Browser tests.
    """

    def setUp(self):
        super(BrowserTest, self).setUp()
        self.browser = z2.Browser(self.portal)
        self.browser.handleErrors = False

    def browser_login(self, user, password):
        login(self.portal, user)
        self.browser.open(self.portal.absolute_url() + '/logout')
        self.browser.open(self.portal.absolute_url() + '/login_form')
        self.browser.getControl(name='__ac_name').value = user
        self.browser.getControl(name='__ac_password').value = password
        self.browser.getControl(name='submit').click()


class ExampleScheduleTestBase(BrowserTest):

    def setUp(self):
        super(ExampleScheduleTestBase, self).setUp()

        # only keep simple tasks
        self.schedule_config = self.portal.config.test_scheduleconfig
        self.empty_schedule_config = self.portal.config.empty_scheduleconfig
        self.task_config = self.schedule_config.test_taskconfig
        self.empty_task_container = self.portal.test_empty_taskcontainer
        self.task_container = self.portal.test_taskcontainer
        self.task = self.task_container.TASK_test_taskconfig

        # commit to save the setup in the tests.
        transaction.commit()

        self.browser_login(TEST_USER_NAME, TEST_USER_PASSWORD)


class ExampleScheduleIntegrationTestCase(ExampleScheduleTestBase):

    layer = EXAMPLE_SCHEDULE_INTEGRATION


class ExampleScheduleFunctionalTestCase(ExampleScheduleTestBase):

    layer = EXAMPLE_SCHEDULE_FUNCTIONAL

    def tearDown(self):
        """
        Unregister the adapters here since the zope instance never shut downs.
        """
        unsubscribe_task_configs_for_content_type(self.task_config, None)

        api.content.delete(self.task_container)
        api.content.delete(self.empty_task_container)
        api.content.delete(self.portal.config)
        zope_registration._registered_sites = set([])

        transaction.commit()

        super(ExampleScheduleTestBase, self).tearDown()


class MacrotaskScheduleLayer(ScheduleLayer):

    def setUpPloneSite(self, portal):
        super(MacrotaskScheduleLayer, self).setUpPloneSite(portal)

        applyProfile(portal, 'Products.CMFPlone:dependencies')
        # ponr skin for robot tests
        applyProfile(portal, 'plonetheme.sunburst:default')
        applyProfile(portal, 'imio.schedule:testing')

        # delete simple tasks
        api.content.delete(
            portal.config.test_scheduleconfig.test_taskconfig
        )
        api.content.delete(
            portal.test_taskcontainer.TASK_test_taskconfig
        )


MACROTASK_SCHEDULE_FIXTURE = MacrotaskScheduleLayer(
    name='MACROTASK_SCHEDULE_FIXTURE'
)

MACROTASK_SCHEDULE_INTEGRATION = IntegrationTesting(
    bases=(MACROTASK_SCHEDULE_FIXTURE,),
    name='MACROTASK_SCHEDULE_INTEGRATION'
)


MACROTASK_SCHEDULE_FUNCTIONAL = FunctionalTesting(
    bases=(MACROTASK_SCHEDULE_FIXTURE,),
    name='MACROTASK_SCHEDULE_FUNCTIONAL'
)


class MacroTaskScheduleTestBase(BrowserTest):

    def setUp(self):
        super(MacroTaskScheduleTestBase, self).setUp()

        # recreate test objects
        self.portal.portal_workflow.setDefaultChain("simple_publication_workflow")

        # only keep macro tasks
        self.schedule_config = self.portal.config.test_scheduleconfig
        self.macrotask_config = self.schedule_config.test_macrotaskconfig
        self.subtask_config = self.macrotask_config.test_subtaskconfig
        self.empty_task_container = self.portal.test_empty_taskcontainer
        self.task_container = self.portal.test_taskcontainer
        self.macro_task = self.task_container.TASK_test_macrotaskconfig
        self.sub_task = self.macro_task.TASK_test_subtaskconfig

        # commit to save the setup in the tests.
        transaction.commit()

        self.browser_login(TEST_USER_NAME, TEST_USER_PASSWORD)


class MacroTaskScheduleIntegrationTestCase(MacroTaskScheduleTestBase):

    layer = MACROTASK_SCHEDULE_INTEGRATION


class MacroTaskScheduleFunctionalTestCase(MacroTaskScheduleTestBase):

    layer = MACROTASK_SCHEDULE_FUNCTIONAL

    def tearDown(self):
        """
        Unregister the adapters here since the zope instance never shut downs.
        """
        unsubscribe_task_configs_for_content_type(self.macrotask_config, None)
        unsubscribe_task_configs_for_content_type(self.subtask_config, None)

        api.content.delete(self.task_container)
        api.content.delete(self.empty_task_container)
        api.content.delete(self.portal.config)

        transaction.commit()

        super(MacroTaskScheduleTestBase, self).tearDown()



# -*- coding: utf-8 -*-

from imio.schedule import _
from imio.schedule.interfaces import ICalculationDelay
from imio.schedule.interfaces import ICreationCondition
from imio.schedule.interfaces import IDefaultTaskGroup
from imio.schedule.interfaces import IDefaultTaskUser
from imio.schedule.interfaces import IEndCondition
from imio.schedule.interfaces import IFreezeCondition
from imio.schedule.interfaces import IMacroTaskCreationCondition
from imio.schedule.interfaces import IMacroTaskEndCondition
from imio.schedule.interfaces import IMacroTaskStartCondition
from imio.schedule.interfaces import IMacroTaskStartDate
from imio.schedule.interfaces import IRecurrenceCondition
from imio.schedule.interfaces import IScheduledContentTypeVocabulary
from imio.schedule.interfaces import IStartCondition
from imio.schedule.interfaces import IStartDate
from imio.schedule.interfaces import ITaskLogic
from imio.schedule.interfaces import ITaskMarkerInterface
from imio.schedule.interfaces import IThawCondition
from imio.schedule.utils import interface_to_tuple
from imio.schedule.utils import dict_list_2_vocabulary

from plone import api

from plone.principalsource.source import PrincipalSource

from Products.CMFPlone import PloneMessageFactory
from Products.CMFPlone.i18nl10n import utranslate

from zope.component import getAdapter
from zope.component import getGlobalSiteManager
from zope.i18n import translate
from zope.interface import implements
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


def unify_vocabularies(vocabularies):
    """Remove duplicates from a vocabulary list"""
    dict = {}
    unified = []
    for vocabularie in vocabularies:
        if vocabularie.value in dict:
            dict[vocabularie.value] += 1
        else:
            dict[vocabularie.value] = 1
            unified.append(vocabularie)
    return unified


class BaseVocabularyFactory(object):
    """
    Base class for vocabulary factories.
    """

    def get_portal_type(self, context):
        """
        Return the portal_type of the (future?) context.
        """
        request = context.REQUEST
        try:
            add_view = request.get('PUBLISHED', request['PARENTS'][-1])
            if hasattr(add_view, 'form_instance'):
                form = add_view.form_instance
                portal_type = form.portal_type
            else:
                form = add_view.context.form_instance
                portal_type = form.portal_type
        except:
            portal_type = context.portal_type

        return portal_type

    def get_fti(self, context):
        """
        Return the fti of the (future?) context.
        """
        portal_type = self.get_portal_type(context)
        portal_types = api.portal.get_tool('portal_types')
        fti = portal_types.getTypeInfo(portal_type)
        return fti


class ScheduledContentTypeVocabularyFactory(BaseVocabularyFactory):
    """
    Vocabulary factory for 'scheduled_contenttype' field.
    Return all the content types that can be associated
    to a task config (=> should implements IScheduledContentType).
    """

    def __call__(self, context):
        """
        Call the adapter vocabulary for the 'scheduled_contenttype' field
        and returns it.
        """
        portal_type = self.get_portal_type(context)
        fti = self.get_fti(context)
        voc_adapter = getAdapter(fti, IScheduledContentTypeVocabulary, portal_type)
        vocabulary = voc_adapter()

        return vocabulary


class ScheduledContentTypeVocabulary(object):
    """
    Adapts a TaskConfig fti to return a specific
    vocabulary for the 'scheduled_contenttype' field.

    Subclass and override allowed_types() and get_message_factory()
    in products using imio.schedule.
    """

    implements(IScheduledContentTypeVocabulary)

    def __init__(self, fti):
        """ """

    def __call__(self):
        """
        Return a vocabulary from an explicit set content types.
        """

        voc_terms = []
        content_types = self.content_types()
        message_factory = self.get_message_factory()

        for portal_type, interface in content_types.iteritems():
            key = self.to_vocabulary_key(portal_type, interface)
            voc_terms.append(
                SimpleTerm(key, key, message_factory(portal_type))
            )

        vocabulary = SimpleVocabulary(voc_terms)

        return vocabulary

    def content_types(self):
        """
        To override.
        Explicitely define here the content types allowed in the field
        'scheduled_contenttype'

        eg:
        from Products.ATContentTypes.interfaces import IATFolder
        return{'Folder': IATFolder}
        """
        return {}

    def get_message_factory(self):
        """
        To override.
        By default return plone MessageFactory.
        """
        return PloneMessageFactory

    def to_vocabulary_key(self, portal_type, interfaces):
        """
        Return the module path of a class.
        """
        if type(interfaces) not in [list, tuple]:
            interfaces = (interfaces,)
        return (portal_type, tuple([interface_to_tuple(i) for i in interfaces]))


class AssignedGroupVocabularyFactory(object):
    """
    Vocabulary factory for 'default_assigned_group' field.
    """

    def __call__(self, context):
        """
        Call the adapter vocabulary for the 'default_assigned_group' field
        Return available groups for a task config.
        """
        gsm = getGlobalSiteManager()
        scheduled_interfaces = context.get_scheduled_interfaces()

        voc_terms = []
        for scheduled_interface in scheduled_interfaces:
            for adapter in gsm.registeredAdapters():
                implements_IGroup = adapter.provided is IDefaultTaskGroup
                specific_enough = adapter.required[0].implementedBy(scheduled_interface) or \
                    issubclass(scheduled_interface, adapter.required[0])
                if implements_IGroup and specific_enough:
                    voc_terms.append(
                        SimpleTerm(adapter.name, adapter.name, _(adapter.name))
                    )

        # enrich the vocabulary with available groups
        for group in api.group.get_groups():
            voc_terms.append(
                SimpleTerm(group.id, group.id, group.getGroupTitleOrName())
            )

        vocabulary = SimpleVocabulary(unify_vocabularies(voc_terms))
        return vocabulary


class AssignedUserVocabularyFactory(object):
    """
    Vocabulary factory for 'default_assigned_user' field.
    """

    def __call__(self, context):
        """
        Call the adapter vocabulary for the 'default_assigned_user' field
        Return available users for a task config.
        """
        gsm = getGlobalSiteManager()
        scheduled_interfaces = context.get_scheduled_interfaces()

        voc_terms = []
        for scheduled_interface in scheduled_interfaces:
            for adapter in gsm.registeredAdapters():
                implements_IUser = adapter.provided is IDefaultTaskUser
                specific_enough = adapter.required[0].implementedBy(scheduled_interface) or \
                    issubclass(scheduled_interface, adapter.required[0])
                if implements_IUser and specific_enough:
                    voc_terms.append(
                        SimpleTerm(adapter.name, adapter.name, _(adapter.name))
                    )

        # enrich the vocabulary with available users
        for user in api.user.get_users():
            voc_terms.append(
                SimpleTerm(user.id, user.id, user.getProperty('fullname') or user.getUserName())
            )

        vocabulary = SimpleVocabulary(sorted(unify_vocabularies(voc_terms), key=lambda term: term.title.decode('utf-8')))
        return vocabulary


class ContainerStateVocabularyFactory(object):
    """
    Vocabulary factory for 'container_state' field.
    """

    def __call__(self, context):
        """
        Call the adapter vocabulary for the 'container_state' field
        and returns it.
        """
        portal_type = context.get_scheduled_portal_type()
        if not portal_type:
            return SimpleVocabulary([])

        wf_tool = api.portal.get_tool('portal_workflow')
        request = api.portal.get().REQUEST

        workfow = wf_tool.get(wf_tool.getChainForPortalType(portal_type)[0])
        voc_terms = [
            SimpleTerm(state_id, state_id, translate(state.title, 'plone', context=request))
            for state_id, state in workfow.states.items()
        ]

        vocabulary = SimpleVocabulary(voc_terms)

        return vocabulary


class TaskMarkerInterfacesVocabulary(object):
    """
    Return available custom Task Marker Interface.
    """

    def __call__(self, context):
        gsm = getGlobalSiteManager()
        interfaces = gsm.getUtilitiesFor(ITaskMarkerInterface)
        items = []

        for interface_name, marker_interface in interfaces:
            items.append(
                SimpleTerm(
                    interface_name,
                    marker_interface.__doc__,
                    utranslate(
                        msgid=marker_interface.__doc__,
                        domain='imio.schedule',
                        context=context,
                        default=marker_interface.__doc__
                    )
                )
            )

        # sort elements by title
        items.sort(lambda a, b: cmp(a.title, b.title))

        return SimpleVocabulary(items)


class BooleanVocabulary(object):
    """
    Return True/False vocabulary.
    """

    def __call__(self, context):
        request = api.portal.get().REQUEST
        items = [
            SimpleTerm('True', 'True', translate('Doable alone', 'imio.schedule', context=request)),
            SimpleTerm('False', 'False', translate('Subtasks dependencies', 'imio.schedule', context=request)),
        ]

        return SimpleVocabulary(items)


class TaskLogicVocabularyFactory(object):
    """
    Base class for vocabulary factories listing adapters providing
    some ITaskLogic (sub)interface and adapting a task container
    content type.
    """

    provides_interface = None  # to override in subclass

    def __call__(self, context):
        """
        Look for all the conditions registered for scheduled_contenttype,
        implementing some ITaskLogic (sub)interface and return them as
        a vocabulary.
        """
        gsm = getGlobalSiteManager()
        scheduled_interfaces = context.get_scheduled_interfaces()

        terms = []
        for scheduled_interface in scheduled_interfaces:
            for adapter in gsm.registeredAdapters():
                implements_interface = issubclass(adapter.provided, ITaskLogic) and \
                    issubclass(self.provides_interface, adapter.provided)
                specific_enough = adapter.required[0].implementedBy(scheduled_interface) or \
                    issubclass(scheduled_interface, adapter.required[0])
                if implements_interface and specific_enough:
                    terms.append(
                        SimpleTerm(adapter.name, adapter.name, _(adapter.name))
                    )

        vocabulary = SimpleVocabulary(unify_vocabularies(terms))
        return vocabulary


class CreationConditionVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'creation_conditions' field.
    Return available creation conditions of a task config.
    """

    provides_interface = ICreationCondition


class StartConditionVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'start_conditions' field.
    Return available start conditions of a task config.
    """

    provides_interface = IStartCondition


class EndConditionVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'end_conditions' field.
    Return available end conditions of a task config.
    """

    provides_interface = IEndCondition


class FreezeConditionVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'end_conditions' field.
    Return available end conditions of a task config.
    """

    provides_interface = IFreezeCondition


class ThawConditionVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'end_conditions' field.
    Return available end conditions of a task config.
    """

    provides_interface = IThawCondition


class StartDateVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'start_date' field.
    Return start date of a task config.
    """

    provides_interface = IStartDate


class RecurrenceConditionVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'recurrence_conditions' field.
    Return recurrence conditions for a task config.
    """
    provides_interface = IRecurrenceCondition


class CalculationDelayVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'calculation_delay' field.
    Return calculation delay methods for a task config.
    """
    provides_interface = ICalculationDelay


class MacroTaskCreationConditionVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'creation_conditions' field.
    Return available creation conditions of a macro task config.
    """

    provides_interface = IMacroTaskCreationCondition


class MacroTaskStartConditionVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'start_conditions' field.
    Return available start conditions of a macro task config.
    """

    provides_interface = IMacroTaskStartCondition


class MacroTaskEndConditionVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'end_conditions' field.
    Return available end conditions of a macro task config.
    """

    provides_interface = IMacroTaskEndCondition


class MacroTaskStartDateVocabularyFactory(TaskLogicVocabularyFactory):
    """
    Vocabulary factory for 'start_date' field.
    Return start date of a macro task config.
    """

    provides_interface = IMacroTaskStartDate


class LogicalOperatorVocabularyFactory(BaseVocabularyFactory):

    def __call__(self, context):
        return dict_list_2_vocabulary([
            {'AND': _(u'and')},
            {'OR': _(u'or')},
        ])


class WeekDaysRoundingVocabulary(object):
    """
    """

    def __call__(self, context):
        raw_terms = [
            ('0', '0', _('None')),
            ('1', '1', _('Next Monday')),
            ('2', '2', _('Next Tuesday')),
            ('3', '3', _('Next Wednesday')),
            ('4', '4', _('Next Thursday')),
            ('5', '5', _('Next Friday')),
            ('6', '6', _('Next Saturday')),
            ('7', '7', _('Next Sunday')),
            ('-1', '-1', _('Previous Monday')),
            ('-2', '-2', _('Previous Tuesday')),
            ('-3', '-3', _('Previous Wednesday')),
            ('-4', '-4', _('Previous Thursday')),
            ('-5', '-5', _('Previous Friday')),
            ('-6', '-6', _('Previous Saturday')),
            ('-7', '-7', _('Previous Sunday')),
        ]
        voc_terms = [SimpleTerm(*term) for term in raw_terms]
        vocabulary = SimpleVocabulary(voc_terms)

        return vocabulary


class TaskOwnerSource(PrincipalSource):

    def __init__(self, context):
        super(TaskOwnerSource, self).__init__(context)

    def _search(self, id=None, exact_match=True):
        users = api.user.get_users(self.context.assigned_group)
        if id is not None:
            return sorted([{'id': u.id} for u in users if u.id == id], key=lambda u_id: u_id['id'])
        return sorted([{'id': u.id} for u in users], key=lambda u_id: u_id['id'])


class TaskOwnerSourceBinder(object):
    """Bind the principal source with either users or groups
    """
    implements(IContextSourceBinder)

    def __call__(self, context):
        return TaskOwnerSource(context)


TaskOwnerVocabulary = TaskOwnerSourceBinder()



# -*- coding: utf-8 -*-

from Acquisition import aq_base
from DateTime import DateTime
from datetime import datetime
from dateutil.relativedelta import relativedelta

from imio.schedule.config import DONE
from imio.schedule.content.delay import CalculationDefaultDelay
from imio.schedule.content.task import AutomatedMacroTask
from imio.schedule.content.task import AutomatedTask
from imio.schedule.testing import ExampleScheduleIntegrationTestCase
from imio.schedule.testing import MacroTaskScheduleIntegrationTestCase
from imio.schedule.testing import TEST_INSTALL_INTEGRATION

from mock import Mock
from plone import api

from zope.annotation import IAnnotations

import unittest


class TestTaskConfig(unittest.TestCase):
    """
    Test TaskConfig content type.
    """

    layer = TEST_INSTALL_INTEGRATION

    def test_TaskConfig_portal_type_is_registered(self):
        portal_types = api.portal.get_tool('portal_types')
        registered_types = portal_types.listContentTypes()
        self.assertTrue('TaskConfig' in registered_types)


class TestTaskConfigFields(ExampleScheduleIntegrationTestCase):
    """
    Test schema fields declaration.
    """

    def test_class_registration(self):
        """
        Check if the class of the content type TaskConfig is the
        correct one.
        """
        from imio.schedule.content.task_config import TaskConfig
        self.assertTrue(self.task_config.__class__ == TaskConfig)

    def test_schema_registration(self):
        """
        Check if the schema Interface of the content type TaskConfig is the
        correct one.
        """
        portal_types = api.portal.get_tool('portal_types')
        taskconfig_type = portal_types.get(self.task_config.portal_type)
        self.assertTrue('ITaskConfig' in taskconfig_type.schema)

    def test_enabled_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'enabled'))

    def test_enabled_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'enabled' is not displayed"
        self.assertTrue('id="form-widgets-enabled"' in contents, msg)
        msg = "field 'enabled' is not translated"
        self.assertTrue('Activé' in contents, msg)

    def test_enabled_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'enabled' is not editable"
        self.assertTrue('Activé' in contents, msg)

    def test_default_assigned_user_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'default_assigned_user'))

    def test_default_assigned_user_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'default_assigned_user' is not displayed"
        self.assertTrue('id="form-widgets-default_assigned_user"' in contents, msg)
        msg = "field 'default_assigned_user' is not translated"
        self.assertTrue('Responsable de la tâche' in contents, msg)

    def test_default_assigned_user_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'default_assigned_user' is not editable"
        self.assertTrue('Responsable de la tâche' in contents, msg)

    def test_creation_conditions_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'creation_conditions'))

    def test_creation_conditions_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'creation_conditions' is not displayed"
        self.assertTrue('id="form-widgets-creation_conditions"' in contents, msg)
        msg = "field 'creation_conditions' is not translated"
        self.assertTrue('Conditions de création' in contents, msg)

    def test_creation_conditions_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'creation_conditions' is not editable"
        self.assertTrue('Conditions de création' in contents, msg)

    def test_creation_state_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'creation_state'))

    def test_creation_state_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'creation_state' is not displayed"
        self.assertTrue('id="form-widgets-creation_state"' in contents, msg)
        msg = "field 'creation_state' is not translated"
        self.assertTrue('État de création de la tâche' in contents, msg)

    def test_creation_state_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'creation_state' is not editable"
        self.assertTrue('État de création de la tâche' in contents, msg)

    def test_start_conditions_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'start_conditions'))

    def test_start_conditions_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'start_conditions' is not displayed"
        self.assertTrue('id="form-widgets-start_conditions"' in contents, msg)
        msg = "field 'start_conditions' is not translated"
        self.assertTrue('Conditions de création' in contents, msg)

    def test_start_conditions_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'start_conditions' is not editable"
        self.assertTrue('Conditions de création' in contents, msg)

    def test_starting_states_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'starting_states'))

    def test_starting_states_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'starting_states' is not displayed"
        self.assertTrue('id="form-widgets-starting_states"' in contents, msg)
        msg = "field 'starting_states' is not translated"
        self.assertTrue('État(s) de démarrage de la tâche' in contents, msg)

    def test_starting_states_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'starting_states' is not editable"
        self.assertTrue('État(s) de démarrage de la tâche' in contents, msg)

    def test_end_conditions_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'end_conditions'))

    def test_end_conditions_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'end_conditions' is not displayed"
        self.assertTrue('id="form-widgets-end_conditions"' in contents, msg)
        msg = "field 'end_conditions' is not translated"
        self.assertTrue('Conditions de clôture' in contents, msg)

    def test_end_conditions_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'end_conditions' is not editable"
        self.assertTrue('Conditions de clôture' in contents, msg)

    def test_ending_states_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'ending_states'))

    def test_ending_states_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'ending_states' is not displayed"
        self.assertTrue('id="form-widgets-ending_states"' in contents, msg)
        msg = "field 'ending_states' is not translated"
        self.assertTrue('État(s) de clôture de la tâche' in contents, msg)

    def test_ending_states_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'ending_states' is not editable"
        self.assertTrue('État(s) de clôture de la tâche' in contents, msg)

    def test_freeze_conditions_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'freeze_conditions'))

    def test_freeze_conditions_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'freeze_conditions' is not displayed"
        self.assertTrue('id="form-widgets-freeze_conditions"' in contents, msg)
        msg = "field 'freeze_conditions' is not translated"
        self.assertTrue('Conditions de gel' in contents, msg)

    def test_freeze_conditions_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'freeze_conditions' is not editable"
        self.assertTrue('Conditions de gel' in contents, msg)

    def test_freeze_states_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'freeze_states'))

    def test_freeze_states_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'freeze_states' is not displayed"
        self.assertTrue('id="form-widgets-freeze_states"' in contents, msg)
        msg = "field 'freeze_states' is not translated"
        self.assertTrue('État(s) de gel de la tâche' in contents, msg)

    def test_freeze_states_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'freeze_states' is not editable"
        self.assertTrue('État(s) de gel de la tâche' in contents, msg)

    def test_thaw_conditions_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'thaw_conditions'))

    def test_thaw_conditions_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'thaw_conditions' is not displayed"
        self.assertTrue('id="form-widgets-thaw_conditions"' in contents, msg)
        msg = "field 'thaw_conditions' is not translated"
        self.assertTrue('Conditions de dégel' in contents, msg)

    def test_thaw_conditions_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'thaw_conditions' is not editable"
        self.assertTrue('Conditions de dégel' in contents, msg)

    def test_thaw_states_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'thaw_states'))

    def test_thaw_states_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'thaw_states' is not displayed"
        self.assertTrue('id="form-widgets-thaw_states"' in contents, msg)
        msg = "field 'thaw_states' is not translated"
        self.assertTrue('État(s) de dégel de la tâche' in contents, msg)

    def test_thaw_states_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'thaw_states' is not editable"
        self.assertTrue('État(s) de dégel de la tâche' in contents, msg)

    def test_start_date_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'start_date'))

    def test_start_date_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'start_date' is not displayed"
        self.assertTrue('id="form-widgets-start_date"' in contents, msg)
        msg = "field 'start_date' is not translated"
        self.assertTrue('Date de départ' in contents, msg)

    def test_start_date_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'start_date' is not editable"
        self.assertTrue('Date de départ' in contents, msg)

    def test_calculation_delay_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'additional_delay'))

    def test_calculation_delay_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'calculation_delay' is not displayed"
        self.assertTrue('id="form-widgets-calculation_delay"' in contents, msg)
        msg = "field 'calculation_delay' is not translated"
        self.assertTrue('Calcul du délai' in contents, msg)

    def test_calculation_delay_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'calculation_delay' is not editable"
        self.assertTrue('Calcul du délai' in contents, msg)

    def test_additional_delay_attribute(self):
        task_config = aq_base(self.task_config)
        self.assertTrue(hasattr(task_config, 'additional_delay'))

    def test_additional_delay_field_display(self):
        self.browser.open(self.task_config.absolute_url())
        contents = self.browser.contents
        msg = "field 'additional_delay' is not displayed"
        self.assertTrue('id="form-widgets-additional_delay"' in contents, msg)
        msg = "field 'additional_delay' is not translated"
        self.assertTrue('Délai supplémentaire' in contents, msg)

    def test_additional_delay_field_edit(self):
        self.browser.open(self.task_config.absolute_url() + '/edit')
        contents = self.browser.contents
        msg = "field 'additional_delay' is not editable"
        self.assertTrue('Délai supplémentaire' in contents, msg)


class TestTaskConfigMethodsIntegration(ExampleScheduleIntegrationTestCase):
    """
    Test TaskConfig methods.
    """

    def setUp(self):
        super(TestTaskConfigMethodsIntegration, self).setUp()
        self._evaluate_one_condition = self.task_config.evaluate_one_condition
        self._evaluate_conditions = self.task_config.evaluate_conditions
        self._match_recurrence_states = self.task_config.match_recurrence_states
        self._create_task = self.task_config.create_task
        self._api_get_state = api.content.get_state
        self._adapter_computed_due_date = CalculationDefaultDelay.compute_due_date

    def tearDown(self):
        self.task_config.evaluate_one_condition = self._evaluate_one_condition
        self.task_config.evaluate_conditions = self._evaluate_conditions
        self.task_config.match_recurrence_states = self._match_recurrence_states
        self.task_config.create_task = self._create_task
        api.content.get_state = self._api_get_state
        CalculationDefaultDelay.compute_due_date = self._adapter_computed_due_date
        super(TestTaskConfigMethodsIntegration, self).tearDown()

    def test_get_task_type(self):
        """
        Should return 'AutomatedTask'
        """
        task_type = self.task_config.get_task_type()
        expected_type = 'AutomatedTask'
        self.assertEquals(task_type, expected_type)

    def test_is_main_taskconfig(self):
        """
        Tells wheter a task config is a subtaskconfig or not
        """
        msg = "this simple task config should be considered as a main task config"
        self.assertTrue(self.task_config.is_main_taskconfig(), msg)

    def test_get_schedule_config(self):
        """
        Should return the parent schedule config.
        """
        config = self.task_config.get_schedule_config()
        expected_config = self.schedule_config
        self.assertEquals(config, expected_config)

    def test_get_scheduled_portal_type(self):
        """
        Sould return the portal_type of the content type selected on the field
        'scheduled_contenttype' of the parent ScheduleConfig.
        """
        portal_type = self.task_config.get_scheduled_portal_type()
        expected_type = 'Folder'
        self.assertEquals(portal_type, expected_type)

    def test_get_scheduled_interfaces(self):
        """
        Should return the Interface (or a class) of the content type selected
        on the field 'scheduled_contenttype' of the parent ScheduleConfig.
        """
        from Products.ATContentTypes.interfaces import IATFolder

        type_interface = self.task_config.get_scheduled_interfaces()
        expected_interface = (IATFolder,)
        self.assertEquals(type_interface, expected_interface)

    def test_user_to_assign(self):
        """
        Should return a user to assign on a task.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        user = task_config.user_to_assign(task_container, task)
        expected_user = 'test-user'
        msg = "should have return '{}' user id".format(expected_user)
        self.assertEquals(user, expected_user, msg)

    def test_get_task_instances(self):
        """
        Should return AutomatedTask brains in a container created from a given TaskConfig.
        """
        task_config = self.task_config

        root = self.portal
        tasks = task_config.get_task_instances(root)
        msg = "Should have found at least one AutomatedTask"
        self.assertEquals(tasks, [self.task], msg)

        root = self.empty_task_container
        tasks = task_config.get_task_instances(root)
        msg = "Should not have found any AutomatedTask"
        self.assertEquals(tasks, [], msg)

    def test_get_task(self):
        """
        Should return the unique Task of task_container created from
        this TaskConfig.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        task_found = task_config.get_task(task_container)
        self.assertEquals(task_found, task)
        # round trip test
        msg = "The TaskConfig of the task found should be the original one"
        self.assertEquals(task_found.get_task_config(), task_config, msg)

    def test_get_created_task(self):
        """
        Should return the unique Task of task_container created from
        this TaskConfig if it is in the state 'created'.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        task_found = task_config.get_created_task(task_container)
        self.assertEquals(task_found, task)

    def test_get_started_task(self):
        """
        Should return the unique Task of task_container created from
        this TaskConfig if it is in the state 'created'.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        # so far nothing should be found
        task_found = task_config.get_started_task(task_container)
        self.assertFalse(task_found)

        # start the task
        api.content.transition(obj=task, transition='do_to_assign')

        task_found = task_config.get_started_task(task_container)
        msg = "should have found the started task"
        self.assertEquals(task_found, task, msg)

    def test_get_open_task(self):
        """
        Should return the unique Task of task_container created from
        this TaskConfig if it is in the state 'created' or 'to_do'.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        task_found = task_config.get_open_task(task_container)
        msg = "should have found the created task"
        self.assertEquals(task_found, task, msg)

        # start the task
        api.content.transition(obj=task, transition='do_to_assign')

        task_found = task_config.get_open_task(task_container)
        msg = "should have found the started task"
        self.assertEquals(task_found, task, msg)

        # close the task
        task_config.end_task(task)

        task_found = task_config.get_open_task(task_container)
        msg = "should not have found the closed task"
        self.assertFalse(task_found, msg)

    def test_get_closed_tasks(self):
        """
        Should return all the closed Task of task_container created
        from this TaskConfig.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        # so far nothing should be found
        task_found = task_config.get_closed_tasks(task_container)
        self.assertFalse(task_found)

        # close the task
        task_config.end_task(task)

        task_found = task_config.get_closed_tasks(task_container)
        msg = "should have found the closed task"
        self.assertEquals(task_found, [task], msg)

    def test_get_closed_task(self):
        """
        Should return the unique Task of task_container created from
        this TaskConfig if it is in the state 'closed'.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        # so far nothing should be found
        task_found = task_config.get_closed_task(task_container)
        self.assertFalse(task_found)

        # close the task
        task_config.end_task(task)

        task_found = task_config.get_closed_task(task_container)
        msg = "should have found the closed task"
        self.assertEquals(task_found, task, msg)

    def test_get_frozen_task(self):
        """
        Should return the unique Task of task_container created from
        this TaskConfig if it is in the state 'frozen'.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        # so far nothing should be found
        task_found = task_config.get_frozen_task(task_container)
        self.assertFalse(task_found)

        # freeze the task
        task_config.freeze_task(task)

        task_found = task_config.get_frozen_task(task_container)
        msg = "should have found the frozen task"
        self.assertEquals(task_found, task, msg)

    def test_task_already_exists(self):
        """
        Should tell wheter the task container already have a task from
        this TaskConfig.
        """
        task_config = self.task_config

        task_container = self.task_container
        msg = "TaskConfig of the existing task found should be the original task_config"
        self.assertTrue(task_config.task_already_exists(task_container), msg)

        empty_task_container = self.empty_task_container
        msg = "no existing task should have been found on empty container"
        self.assertFalse(task_config.task_already_exists(empty_task_container), msg)

    def test_should_create_task(self):
        """
        Test different cases for the 'should_create_task' method.
        """
        task_config = self.task_config
        task_container = self.task_container
        empty_task_container = self.empty_task_container

        # case of task already existing
        msg = "Task should not be created because it already exists"
        self.assertFalse(task_config.should_create_task(task_container), msg)

        # normal case
        msg = "Task should be created"
        create = task_config.should_create_task(empty_task_container)
        self.assertTrue(create, msg)

        # no creation states given means any state allow task creation
        task_config.creation_state = None
        create = task_config.should_create_task(empty_task_container)
        self.assertTrue(create, msg)

        # disable the task config => task should not be created
        task_config.enabled = False
        msg = "Task should not be created because the creation condition is not matched"
        self.assertFalse(task_config.should_create_task(empty_task_container), msg)

        # set the task_config field 'creation_conditions' with a negative condition
        # => task should not be created
        task_config.enabled = True
        task_config.creation_conditions = [type('condition', (object, ), {
            'condition': 'schedule.negative_creation_condition',
            'operator': 'AND',
        })()]
        msg = "Task should not be created because the creation condition is not matched"
        self.assertFalse(task_config.should_create_task(empty_task_container), msg)

        # set the task_config starting_states field to a state different from
        # the task_container state => task should not be created
        task_config.creation_conditions = [type('condition', (object, ), {
            'condition': 'schedule.test_creation_condition',
            'operator': 'AND',
        })()]
        task_config.creation_state = 'pending'
        msg = "Task should not be created because the creation state does not match container state"
        self.assertFalse(task_config.should_create_task(empty_task_container), msg)

    def test_should_start_task(self):
        """
        Test different cases for the 'should_start_task' method.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        # task container state is different from the states selected on
        # task_config 'starting_states' => task should not start
        msg = "Task should not be started because the starting state does not match container state"
        self.assertFalse(task_config.should_start_task(task_container, task), msg)

        # normal case
        api.content.transition(obj=task_container, transition='submit')
        msg = "Task should be started"
        start = task_config.should_start_task(task_container, task)
        self.assertTrue(start, msg)

        # no starting creation states given means any state allow task start
        task_config.starting_states = None
        start = task_config.should_start_task(task_container, task)
        self.assertTrue(start, msg)

        # set the task_config field 'start_conditions' with a negative condition
        # => task should not start
        task.assigned_user = None
        msg = "Task should not be started because no user is defined on the task"
        self.assertFalse(task_config.should_start_task(task_container, task), msg)

        # set the task_config field 'start_conditions' with a negative condition
        # => task should not start
        task.assigned_user = 'user'
        task_config.start_conditions = [type('condition', (object, ), {
            'condition': 'schedule.negative_start_condition',
            'operator': 'AND',
        })()]
        msg = "Task should not be started because the start condition is not matched"
        self.assertFalse(task_config.should_start_task(task_container, task), msg)

        # set the task_config starting_states field to a state different from
        # the task_container state => task should not start
        task_config.start_conditions = [type('condition', (object, ), {
            'condition': 'schedule.test_start_condition',
            'operator': 'AND',
        })()]
        task_config.starting_states = ('published',)
        msg = "Task should not be started because the starting state does not match container state"
        self.assertFalse(task_config.should_start_task(task_container, task), msg)

    def test_start_conditions_status(self):
        """
        Test different cases for the 'start_conditions_status' method.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        matched_conditions, unmatched_conditions = task_config.start_conditions_status(task_container, task)
        self.assertTrue(matched_conditions == ['schedule.test_start_condition'])
        self.assertTrue(unmatched_conditions == [])

        task_config.start_conditions = [type('condition', (object, ), {
            'condition': 'schedule.negative_start_condition',
            'operator': 'AND',
        })()]
        matched_conditions, unmatched_conditions = task_config.start_conditions_status(task_container, task)
        self.assertTrue(matched_conditions == [])
        self.assertTrue(unmatched_conditions == ['schedule.negative_start_condition'])

    def test_should_end_task(self):
        """
        Test different cases for the 'should_end_task' method.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        # task container state is different from the states selected on
        # task_config 'ending_states' => task should not end
        msg = "Task should not be ended because the ending state does not match container state"
        self.assertFalse(task_config.should_end_task(task_container, task), msg)

        # normal case
        api.content.transition(obj=task_container, transition='publish')
        msg = "Task should be ended"
        end = task_config.should_end_task(task_container, task)
        self.assertTrue(end, msg)

        # no ending creation states given means any state allow task end
        task_config.ending_states = None
        end = task_config.should_end_task(task_container, task)
        self.assertTrue(end, msg)

        # set the task_config field 'end_conditions' with a negative condition
        # => task should not end
        task_config.end_conditions = [type('object', (object, ), {
            'condition': 'schedule.negative_end_condition',
            'operator': 'AND',
        })()]
        msg = "Task should not be ended because the end condition is not matched"
        self.assertFalse(task_config.should_end_task(task_container, task), msg)

        # set the task_config ending_states field to a state different from
        # the task_container state => task should not end
        task_config.ending_states = ('pending',)
        msg = "Task should not be ended because the ending state does not match container state"
        self.assertFalse(task_config.should_end_task(task_container, task), msg)

    def test_end_conditions_status(self):
        """
        Test different cases for the 'end_conditions_status' method.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        matched_conditions, unmatched_conditions = task_config.end_conditions_status(task_container, task)
        self.assertTrue(matched_conditions == ['schedule.test_end_condition'])
        self.assertTrue(unmatched_conditions == [])

        task_config.end_conditions = [type('condition', (object, ), {
            'condition': 'schedule.negative_end_condition',
            'operator': 'AND',
        })()]
        matched_conditions, unmatched_conditions = task_config.end_conditions_status(task_container, task)
        self.assertTrue(matched_conditions == [])
        self.assertTrue(unmatched_conditions == ['schedule.negative_end_condition'])

    def test_should_freeze_task(self):
        """
        Test different cases for the 'should_freeze_task' method.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task
        # prevent ending task during the freeze test
        task_config.ending_states = []

        # task container state is different from the states selected on
        # task_config 'freeze_states' => task should not freeze
        msg = "Task should not be frozen because the freeze state does not match container state"
        self.assertFalse(task_config.should_freeze_task(task_container, task), msg)

        # normal case
        task_config.freeze_states = ['private']
        msg = "Task should be frozen"
        freeze = task_config.should_freeze_task(task_container, task)
        self.assertTrue(freeze, msg)

        # no freeze creation states given means the task should never be frozen
        task_config.freeze_states = None
        freeze = task_config.should_freeze_task(task_container, task)
        self.assertFalse(freeze, msg)

        # set the task_config field 'freeze_conditions' with a negative condition
        # => task should not freeze
        task_config.freeze_states = ['private']
        task_config.freeze_conditions = [type('object', (object, ), {
            'condition': 'schedule.negative_freeze_condition',
            'operator': 'AND',
        })()]
        msg = "Task should not be frozen because the freeze condition is not matched"
        freeze = task_config.should_freeze_task(task_container, task)
        self.assertFalse(freeze, msg)

        # set the task_config freeze_states field to a state different from
        # the task_container state => task should not freeze
        task_config.freeze_conditions = []
        task_config.freeze_states = ('pending',)
        msg = "Task should not be frozen because the freeze state does not match container state"
        freeze = task_config.should_freeze_task(task_container, task)
        self.assertFalse(freeze, msg)

    def test_freeze_conditions_status(self):
        """
        Test different cases for the 'freeze_conditions_status' method.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        matched_conditions, unmatched_conditions = task_config.freeze_conditions_status(task_container, task)
        self.assertTrue(matched_conditions == ['schedule.test_freeze_condition'])
        self.assertTrue(unmatched_conditions == [])

        task_config.freeze_conditions = [type('condition', (object, ), {
            'condition': 'schedule.negative_freeze_condition',
            'operator': 'AND',
        })()]
        matched_conditions, unmatched_conditions = task_config.freeze_conditions_status(task_container, task)
        self.assertTrue(matched_conditions == [])
        self.assertTrue(unmatched_conditions == ['schedule.negative_freeze_condition'])

    def test_should_thaw_task(self):
        """
        Test different cases for the 'should_thaw_task' method.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task
        # freeze task befor thaw tests
        task_config.freeze_task(task)
        self.assertEquals(api.content.get_state(task), 'frozen', 'task should be frozen')

        # task container state is different from the states selected on
        # task_config 'thaw_states' => task should not thaw
        msg = "Task should not be thawed because the thaw state does not match container state"
        self.assertFalse(task_config.should_thaw_task(task_container, task), msg)

        # normal case
        task_config.thaw_states = ['private']
        msg = "Task should be thawed"
        thaw = task_config.should_thaw_task(task_container, task)
        self.assertTrue(thaw, msg)

        # no thaw creation states given means the task should never be thawed
        task_config.thaw_states = None
        thaw = task_config.should_thaw_task(task_container, task)
        self.assertFalse(thaw, msg)

        # set the task_config field 'thaw_conditions' with a negative condition
        # => task should not thaw
        task_config.thaw_states = ['private']
        task_config.thaw_conditions = [type('object', (object, ), {
            'condition': 'schedule.negative_thaw_condition',
            'operator': 'AND',
        })()]
        msg = "Task should not be thawed because the thaw condition is not matched"
        thaw = task_config.should_thaw_task(task_container, task)
        self.assertFalse(thaw, msg)

        # set the task_config thaw_states field to a state different from
        # the task_container state => task should not thaw
        task_config.thaw_conditions = []
        task_config.thaw_states = ('pending',)
        msg = "Task should not be thawed because the thaw state does not match container state"
        thaw = task_config.should_thaw_task(task_container, task)
        self.assertFalse(thaw, msg)

    def test_thaw_conditions_status(self):
        """
        Test different cases for the 'thaw_conditions_status' method.
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task
        # freeze task befor thaw tests
        task_config.freeze_task(task)
        self.assertEquals(api.content.get_state(task), 'frozen', 'task should be frozen')

        matched_conditions, unmatched_conditions = task_config.thaw_conditions_status(task_container, task)
        self.assertTrue(matched_conditions == ['schedule.test_thaw_condition'])
        self.assertTrue(unmatched_conditions == [])

        task_config.thaw_conditions = [type('condition', (object, ), {
            'condition': 'schedule.negative_thaw_condition',
            'operator': 'AND',
        })()]
        matched_conditions, unmatched_conditions = task_config.thaw_conditions_status(task_container, task)
        self.assertTrue(matched_conditions == [])
        self.assertTrue(unmatched_conditions == ['schedule.negative_thaw_condition'])

    def test_create_task(self):
        """
        Should create a task.
        """
        task_container = self.empty_task_container
        task_config = self.task_config

        created_task = task_config.create_task(task_container)

        self.assertTrue(isinstance(created_task, AutomatedTask))

        # should raise TaskAlreadyExists when trying to use same task id
        from imio.schedule.interfaces import TaskAlreadyExists
        kwargs = {'task_container': task_container, 'task_id': created_task.id}
        self.assertRaises(TaskAlreadyExists, task_config.create_task, **kwargs)

    def test_start_task(self):
        """
        Should start a task.
        """
        task_config = self.task_config
        task = self.task

        msg = "task should not be started yet (for the sake of the test..)"
        self.assertEquals(api.content.get_state(task), 'created', msg)

        task_config.start_task(task)

        msg = "task should have been started"
        self.assertEquals(api.content.get_state(task), 'to_do', msg)

    def test_end_task(self):
        """
        Should end a task.
        """
        task_config = self.task_config
        task = self.task

        msg = "task should not be ended yet (for the sake of the test..)"
        self.assertEquals(api.content.get_state(task), 'created', msg)

        task_config.end_task(task)

        msg = "task should have been ended"
        self.assertEquals(api.content.get_state(task), 'closed', msg)

    def test_freeze_task(self):
        """
        Should freeze a task.
        """
        task_config = self.task_config
        task = self.task

        msg = "task should not be frozen yet (for the sake of the test..)"
        self.assertEquals(api.content.get_state(task), 'created', msg)

        task_config.freeze_task(task)

        msg = "task should have been frozen"
        self.assertEquals(api.content.get_state(task), 'frozen', msg)

    def test_thaw_task(self):
        """
        Should thaw a task.
        """
        task_config = self.task_config
        task = self.task

        msg = "task should be in created state"
        self.assertEquals(api.content.get_state(task), 'created', msg)

        # freeze task befor thaw tests
        task_config.freeze_task(task)
        msg = "task should not be thawed yet (for the sake of the test..)"
        self.assertEquals(api.content.get_state(task), 'frozen', msg)

        task_config.thaw_task(task)

        msg = "task should have been thawed and put back in its original state"
        self.assertEquals(api.content.get_state(task), 'created', msg)

    def test_thaw_task_stack_freeze_periods(self):
        """
        Verify that a task accumulate multiple frozen periods correctly.
        Multiple freeze periods should stack.
        """
        task_config = self.task_config
        task = self.task

        msg = "task should be in created state"
        self.assertEquals(api.content.get_state(task), 'created', msg)

        # freeze task befor thaw tests
        task_config.freeze_task(task)
        msg = "task should not be thawed yet (for the sake of the test..)"
        self.assertEquals(api.content.get_state(task), 'frozen', msg)

        # mock a previous freeze period of 5 days and mockput the current freeze date
        # as 10 days earlier to simulate a new 10 days freeze period
        annotations = IAnnotations(task)
        freeze_infos = annotations['imio.schedule.freeze_task']
        freeze_infos['previous_freeze_duration'] = 5
        freeze_infos['freeze_date'] = str(datetime.now().date() + relativedelta(days=-10))
        annotations['imio.schedule.freeze_task'] = freeze_infos

        task_config.thaw_task(task)

        msg = "task should have been thawed and put back in its original state"
        self.assertEquals(api.content.get_state(task), 'created', msg)
        msg = "task freeze duration should be the sum of the two freeze periods: 5 + 10"
        freeze_period = annotations['imio.schedule.freeze_task']['previous_freeze_duration']
        self.assertEquals(freeze_period, 15, msg)

    def test_compute_due_date(self):
        """
        Due date should be the date computed by the adapter of
        start_date field + the value in additional_delay (10)
        + the computed delay (15).
        """
        task_config = self.task_config
        task_container = self.task_container
        task = self.task

        additional_delay = self.task_config.additional_delay
        computed_delay = 15
        CalculationDefaultDelay.calculate_delay = Mock(return_value=computed_delay)

        expected_date = task_container.creation_date + computed_delay + additional_delay
        expected_date = expected_date.asdatetime().date()

        due_date = task_config.compute_due_date(task_container, task)
        self.assertEquals(due_date, expected_date)

    def test_compute_due_date_working_days(self):
        """
        Due date should handle the working days
        start date (2017-03-01) + 10 additional days = 2017-03-11
        + 4 weekend days = 2017-03-15
        """
        task_config = self.task_config
        task_container = self.task_container
        task_container.creation_date = DateTime(2017, 3, 1)
        task = self.task

        CalculationDefaultDelay.calculate_delay = Mock(return_value=0)

        task_config.additional_delay_type = 'working_days'
        due_date = task_config.compute_due_date(task_container, task)
        self.assertEquals(due_date, datetime(2017, 3, 15).date())

    def test_compute_due_date_holidays(self):
        """
        Due date whould handle the holidays and working days
        start date (2017-04-14) + 10 additional days = 2017-04-25
        + easter day (2017-04-17) + labour day (2017-05-01) + 6 weekend days
        = 2017-05-02
        """
        task_config = self.task_config
        task_container = self.task_container
        task_container.creation_date = DateTime(2017, 4, 14)
        task = self.task

        CalculationDefaultDelay.calculate_delay = Mock(return_value=0)

        task_config.additional_delay_type = 'working_days'
        due_date = task_config.compute_due_date(task_container, task)
        self.assertEquals(due_date, datetime(2017, 5, 2).date())

    def test_compute_due_date_with_freeze_period(self):
        """
        Test 'compute_due_date' method for a task who had been frozen.
        start date (2017-04-01) + 10 additional days = 2017-04-11
        + 10 days of frozen task
        """
        task_config = self.task_config
        task_container = self.task_container
        task_container.creation_date = DateTime(2017, 4, 1)
        task = self.task

        CalculationDefaultDelay.calculate_delay = Mock(return_value=0)

        annotations = IAnnotations(task)
        annotations['imio.schedule.freeze_task'] = {
            'freeze_date': None,
            'previous_state': task.get_state(),
            'previous_freeze_duration': 10
        }

        due_date = task_config.compute_due_date(task_container, task)
        self.assertEquals(due_date, datetime(2017, 4, 21).date())

    def test_evaluate_conditions_and(self):
        """
        Evaluate conditions with AND operator
        """
        conditions = [
            type('condition', (object, ), {'condition': 1, 'operator': 'AND'})(),
            type('condition', (object, ), {'condition': 2, 'operator': 'AND'})(),
        ]
        self.task_config.evaluate_one_condition = Mock(return_value=True)
        result = self.task_config.evaluate_conditions(conditions, None, None)
        self.assertTrue(result)

        self.task_config.evaluate_one_condition = Mock(side_effect=[True, False])
        result = self.task_config.evaluate_conditions(conditions, None, None)
        self.assertFalse(result)

    def test_evaluate_conditions_or(self):
        """
        Evaluate conditions with OR operator
        """
        conditions = [
            type('condition', (object, ), {'condition': 1, 'operator': 'OR'})(),
            type('condition', (object, ), {'condition': 2, 'operator': 'OR'})(),
        ]
        self.task_config.evaluate_one_condition = Mock(return_value=True)
        result = self.task_config.evaluate_conditions(conditions, None, None)
        self.assertTrue(result)

        self.task_config.evaluate_one_condition = Mock(side_effect=[True, False])
        result = self.task_config.evaluate_conditions(conditions, None, None)
        self.assertTrue(result)

        self.task_config.evaluate_one_condition = Mock(side_effect=[False, True])
        result = self.task_config.evaluate_conditions(conditions, None, None)
        self.assertTrue(result)

        self.task_config.evaluate_one_condition = Mock(side_effect=[False, False])
        result = self.task_config.evaluate_conditions(conditions, None, None)
        self.assertFalse(result)

    def test_evaluate_conditions_and_or(self):
        """
        Evaluate conditions with AND and OR operators
        """
        conditions = [
            type('condition', (object, ), {'condition': 1, 'operator': 'OR'})(),
            type('condition', (object, ), {'condition': 2, 'operator': 'AND'})(),
            type('condition', (object, ), {'condition': 3, 'operator': 'AND'})(),
        ]
        self.task_config.evaluate_one_condition = Mock(side_effect=[True, False, False])
        result = self.task_config.evaluate_conditions(conditions, None, None)
        self.assertTrue(result)

        self.task_config.evaluate_one_condition = Mock(side_effect=[False, False, True])
        result = self.task_config.evaluate_conditions(conditions, None, None)
        self.assertFalse(result)

        self.task_config.evaluate_one_condition = Mock(side_effect=[False, True, True])
        result = self.task_config.evaluate_conditions(conditions, None, None)
        self.assertTrue(result)

        conditions = [
            type('condition', (object, ), {'condition': 1, 'operator': 'OR'})(),
            type('condition', (object, ), {'condition': 2, 'operator': 'AND'})(),
            type('condition', (object, ), {'condition': 3, 'operator': 'OR'})(),
        ]
        self.task_config.evaluate_one_condition = Mock(side_effect=[False, False, True])
        result = self.task_config.evaluate_conditions(conditions, None, None)
        self.assertFalse(result)

    def test_should_recurred(self):
        """
        Test different cases for the 'should_recurred' method
        """
        task_config = self.task_config
        task_config.activate_recurrency = False
        task_config.recurrence_conditions = None
        task_config.match_recurrence_states = Mock(return_value=False)
        task_config.evaluate_conditions = Mock(return_value=False)
        self.assertFalse(task_config.should_recurred(None))

        task_config.match_recurrence_states = Mock(return_value=True)
        self.assertFalse(task_config.should_recurred(None))

        task_config.activate_recurrency = True
        self.assertTrue(task_config.should_recurred(None))

        task_config.recurrence_conditions = ['foo']
        task_config.evaluate_conditions = Mock(return_value=True)
        self.assertTrue(task_config.should_recurred(None))

    def test_match_recurrence_states(self):
        """
        Test method 'match_recurrence_states'
        """
        self.task_config.recurrence_states = []
        self.assertTrue(self.task_config.match_recurrence_states(None))

        self.task_config.recurrence_states = ['foo']
        api.content.get_state = Mock(return_value='foo')
        self.assertTrue(self.task_config.match_recurrence_states(None))

        self.task_config.recurrence_states = ['bar']
        self.assertFalse(self.task_config.match_recurrence_states(None))

    def test_create_recurring_task(self):
        """
        Test different cases for the 'create_recurring_task' method
        """
        container = type('container', (dict, ), {})()
        container['TASK_test_taskconfig'] = None
        api.content.get_state = Mock(return_value='foo')
        container.objectIds = Mock(return_value=['TASK_test_taskconfig'])
        self.assertIsNone(self.task_config.create_recurring_task(
            container,
            creation_place=container,
        ))

        container.objectIds = Mock(return_value=[])
        self.task_config.create_task = Mock(return_value=True)
        self.assertTrue(self.task_config.create_recurring_task(
            container,
            creation_place=container,
        ))

        container.objectIds = Mock(return_value=['TASK_test_taskconfig'])
        api.content.get_state = Mock(return_value='closed')
        self.assertTrue(self.task_config.create_recurring_task(container))


class TestMacroTaskConfig(unittest.TestCase):
    """
    Test MacroTaskConfig content type.
    """

    layer = TEST_INSTALL_INTEGRATION

    def test_MacroTaskConfig_portal_type_is_registered(self):
        portal_types = api.portal.get_tool('portal_types')
        registered_types = portal_types.listContentTypes()
        self.assertTrue('MacroTaskConfig' in registered_types)


class TestMacroTaskConfigMethodsIntegration(MacroTaskScheduleIntegrationTestCase):
    """
    Test MacroTaskConfig methods.
    """

    def setUp(self):
        super(TestMacroTaskConfigMethodsIntegration, self).setUp()
        self._mt_match_recurrence_states = self.macrotask_config.match_recurrence_states
        self._mt_evaluate_conditions = self.macrotask_config.evaluate_conditions
        self._st_match_recurrence_states = self.subtask_config.match_recurrence_states
        self._st_evaluate_conditions = self.subtask_config.evaluate_conditions

    def tearDown(self):
        self.macrotask_config.match_recurrence_states = self._mt_match_recurrence_states
        self.macrotask_config.evaluate_conditions = self._mt_evaluate_conditions
        self.subtask_config.match_recurrence_states = self._st_match_recurrence_states
        self.subtask_config.evaluate_conditions = self._st_evaluate_conditions
        super(TestMacroTaskConfigMethodsIntegration, self).tearDown()

    def test_MacroTasConfigk_inherits_from_TaskConfig(self):
        """
        MacroTaskConfig implements ITaskConfig and should implements all
        the methods defined in TaskConfig.
        As long MacroTaskConfig inherits from BaseTaskConfig, we test the inherited
        methods only in TaskConfig test cases to avoid test code duplication.
        """
        from imio.schedule.content.task_config import MacroTaskConfig
        from imio.schedule.content.task_config import BaseTaskConfig
        msg = "MacroTaskConfig should inherits BaseTaskConfig"
        self.assertTrue(issubclass(MacroTaskConfig, BaseTaskConfig), msg)

    def test_get_task_type(self):
        """
        Should return 'AutomatedMacroTask'
        """
        task_type = self.macrotask_config.get_task_type()
        expected_type = 'AutomatedMacroTask'
        self.assertEquals(task_type, expected_type)

    def test_get_subtask_configs(self):
        """
        Should return all the subtasks configs of a macro task.
        """
        subtasks_configs = self.macrotask_config.get_subtask_configs()
        msg = "sould have return the subtask config"
        self.assertTrue(len(subtasks_configs) == 1, msg)
        self.assertTrue(subtasks_configs[0] == self.subtask_config, msg)

    def test_is_main_taskconfig(self):
        """
        Tells wheter a task config is a subtaskconfig or not
        """
        msg = "this macrotask config should be considered as a main task config"
        self.assertTrue(self.macrotask_config.is_main_taskconfig(), msg)
        msg = "this subtask config should not be considered as a main task config"
        self.assertFalse(self.subtask_config.is_main_taskconfig(), msg)

    def test_create_task(self):
        """
        Should create a macro task and all its subtasks.
        """
        task_container = self.empty_task_container
        task_config = self.macrotask_config

        created_macrotask = task_config.create_task(task_container)
        self.assertTrue(isinstance(created_macrotask, AutomatedMacroTask))

        created_subtask = created_macrotask.objectValues()[0]
        self.assertTrue(isinstance(created_subtask, AutomatedTask))

    def test_should_end_task(self):
        """
        Test different cases for the 'should_end_task' method.
        """
        macrotask_config = self.macrotask_config
        subtask_config = self.subtask_config
        task_container = self.task_container
        macrotask = self.macro_task
        subtask = self.sub_task

        # normal case
        api.content.transition(obj=task_container, transition='publish')

        msg = "SubTask should be ended"
        end = subtask_config.should_end_task(task_container, macrotask)
        self.assertTrue(end, msg)

        msg = "MacroTask should be ended"
        end = macrotask_config.should_end_task(task_container, subtask)
        self.assertTrue(end, msg)

        # re open the subtask => the macro task should not be ended as long
        # the subtask is open

        api.content.transition(obj=subtask, transition='back_in_realized')
        self.assertFalse(subtask.get_status() == DONE)

        msg = "MacroTask should not be ended as long its subtask is open"
        end = macrotask_config.should_end_task(task_container, macrotask)
        self.assertFalse(end, msg)

    def test_should_recurred_macrotask(self):
        """
        Test different cases for the 'should_recurred' method for macro task
        """
        task_config = self.macrotask_config
        task_config.activate_recurrency = False
        task_config.recurrence_conditions = None
        task_config.match_recurrence_states = Mock(return_value=False)
        task_config.evaluate_conditions = Mock(return_value=False)
        self.assertFalse(task_config.should_recurred(None))

        task_config.match_recurrence_states = Mock(return_value=True)
        self.assertFalse(task_config.should_recurred(None))

        task_config.activate_recurrency = True
        self.assertTrue(task_config.should_recurred(None))

        task_config.recurrence_conditions = ['foo']
        task_config.evaluate_conditions = Mock(return_value=True)
        self.assertTrue(task_config.should_recurred(None))

    def test_freeze_macrotask(self):
        macrotask_config = self.macrotask_config
        macrotask = self.macro_task
        subtask = self.sub_task

        # normal case
        macrotask_config.freeze_task(macrotask)

        msg = "MacroTask should be frozen"
        self.assertEquals(macrotask.get_state(), 'frozen', msg)

        msg = "SubTask should be frozen"
        self.assertEquals(subtask.get_state(), 'frozen', msg)

    def test_thaw_macrotask(self):
        macrotask_config = self.macrotask_config
        macrotask = self.macro_task
        subtask = self.sub_task

        original_macrotask_state = macrotask.get_state()
        original_subtask_state = subtask.get_state()
        # freeze task before trying to thaw it
        macrotask_config.freeze_task(macrotask)

        msg = "MacroTask should be frozen"
        self.assertEquals(macrotask.get_state(), 'frozen', msg)
        msg = "SubTask should be frozen"
        self.assertEquals(subtask.get_state(), 'frozen', msg)

        # thaw macro task
        macrotask_config.thaw_task(macrotask)

        msg = "MacroTask should be thawed"
        self.assertEquals(macrotask.get_state(), original_macrotask_state, msg)
        msg = "SubTask should be thawed"
        self.assertEquals(subtask.get_state(), original_subtask_state, msg)



# -*- coding: utf-8 -*-

from imio.schedule.content.condition import CreationCondition
from imio.schedule.content.condition import EndCondition
from imio.schedule.content.condition import FreezeCondition
from imio.schedule.content.condition import StartCondition
from imio.schedule.content.condition import ThawCondition


class TestCreationCondition(CreationCondition):
    """
    Test task start condition.
    """

    def evaluate(self):
        return 'Should start'


class TestNegativeCreationCondition(CreationCondition):
    """
    Test task start condition.
    """

    def evaluate(self):
        return False


class TestStartCondition(StartCondition):
    """
    Test task start condition.
    """

    def evaluate(self):
        return 'Should start'


class TestNegativeStartCondition(StartCondition):
    """
    Test task start condition.
    """

    def evaluate(self):
        return False


class TestEndCondition(EndCondition):
    """
    Test task end condition.
    """

    def evaluate(self):
        return 'Should end'


class TestNegativeEndCondition(EndCondition):
    """
    Test task end condition.
    """

    def evaluate(self):
        return False


class TestFreezeCondition(FreezeCondition):
    """
    Test task freeze condition.
    """

    def evaluate(self):
        return 'Should freeze'


class TestNegativeFreezeCondition(FreezeCondition):
    """
    Test task freeze condition.
    """

    def evaluate(self):
        return False


class TestThawCondition(ThawCondition):
    """
    Test task thaw condition.
    """

    def evaluate(self):
        return 'Should thaw'


class TestNegativeThawCondition(ThawCondition):
    """
    Test task thaw condition.
    """

    def evaluate(self):
        return False
